
from sys import *
from oldnumeric.mlab import *
import gzip
from basics import *
from numpy.core.numerictypes import *
import numpy
from scipy.special import *
from glob import glob
from pgplot import *
from FITS import *
import pickle
import cPickle
import os
from scipy.interpolate import *
from Multipack import bisplev2

if os.environ.has_key('BC03'): bcdir = os.environ['BC03']
else: bcdir = '/Users/kelson/GalMod/bc03'
udir = bcdir + "/models/Padova1994/chabrier/"

def load1prop(u,prop):
   for t2 in ["ABmag","color","lsindx_sed_lick_system"]:
      for t1 in 1+arange(7):
         file = udir+"bc2003_hr_m%d_chab_ssp.%d%s" % (u,t1,t2)
         if os.path.exists(file):
           i = 0; o = open(file,"r")
           l = o.readline()
           if "lick" in t2:
             while "# log-age" not in l: i,l= i+1,o.readline()
           else:
             while "#log-age-yr" not in l: i,l= i+1,o.readline()
           o.close()
           if prop in l.split():
              d = readtable(file,header=i)
              if "lick" in t2: d["log-age-yr"] = d["log-age"]
              return d["log-age-yr"],d[prop]

class BCProperty:
   def __init__(self,prop,kx,ky,l=0,verb=0):
      self.prop = prop
      self.load(l)
      self.fit(kx,ky,verb=verb)
   def load(self,l):
      m = [22,32,42,52,62,72]
      z = [0.0001,0.0004,0.004,0.008,0.02,0.05]
      p = [load1prop(u,self.prop) for u in m]
      a = asarray([x[0] for x in p])
      o = asarray([x[1] for x in p])
      z = asarray([z[j]*ones(a[j].shape,float32) for j in range(len(z))])
      self.z = log10(z/0.02)
      self.a = a
      self.o = o
      if l: self.o = getlog10(self.o)
   def fit(self,kx=5,ky=5,verb=0):
      tx = self.a[0]
      ty = self.z[::,0]
      while len(ty) < 2*ky+2: ty = concatenate([ty[:1],ty,ty[-1:]])
      if verb:
         print tx
         print ty
      self.tck = bisplrep(self.a.ravel(),self.z.ravel(),self.o.ravel(),kx=kx,ky=ky,s=0,\
        task=-1,tx=tx,ty=ty,nxest=2*len(tx),nyest=2*len(ty),quiet=0,full_output=1)
   def interp(self,lage,feh):
      if type(lage) == numpy.ndarray and type(feh) == numpy.ndarray and lage.shape == feh.shape:
        return bisplev2(lage,feh,self.tck[0])
      else:
        return bisplev([lage],[feh],self.tck[0])
   def tau(self,lage,lend,tau,Z,ninc=16,pmag=0,plog=0,pinv=0,emin=0,mage=0):
      if type(lage) == numpy.ndarray:
         v = array([self.tau1(lage[j],lend[j],tau[j],Z[j],ninc=ninc,pmag=pmag,plog=plog,pinv=pinv,emin=emin,mage=mage) for j in range(len(lage))])
      else:
         v = self.tau1(lage,lend,tau,Z,ninc=ninc,pmag=pmag,plog=plog,emin=emin,mage=mage)
      return v
   def tau1(self,lage,lend,ltau,Z,ninc=16,pmag=0,plog=0,pinv=0,mage=0,emin=5):
      start = pow(10.0,lage)
      finish = max([ exp10(lend),exp10(emin)])
      tau = exp10(ltau)
      if type(Z) is StringType: Zu = lambda t: eval(Z)
      elif type(Z) is FloatType: Zu = lambda t: Z
      else: Zu = lambda t: Z
      if finish<start and ltau:
#         if (start-finish) > 5*tau: finish = start - 5*tau
#         dt = min([tau/(ninc/2),(start-finish)/ninc])
#         ninc2 = int((start-finish)/dt+1)
         dt = (log10(start)-log10(finish))/ninc
         times = exp10(arange(log10(finish),log10(start)+dt/100,dt,float32))
         ltimes = log(times)
         dt = (times[2:]-times[:-2])/2.
         dt = concatenate([dt[:1],dt,dt[-1:]])
         dlt = (ltimes[2:]-ltimes[:-2])/2.
         dlt = concatenate([dlt[:1],dlt,dlt[-1:]])
         dt = times*dlt
         SF = lambda t: exp(-(start-t)/tau)
         SFs = array([SF(moment) for moment in times])
         SFtot = sum(dt*SFs)
         SFs = dt*SFs/SFtot
         Ps = array([self.interp(log10(moment),Zu(moment)) for moment in times])
         if mage:
            if pmag:
              vals = exp10(sum(SFs*exp10(-0.4*Ps)*log10(times)) / sum(SFs*exp10(-0.4*Ps)))
            elif plog:
              vals = exp10(sum(SFs*exp10(Ps)*log10(times)) / sum(SFs*exp10(Ps)))
            else:
              vals = exp10(sum(SFs*log10(times))/sum(SFs))
         elif pmag:
            vals =-2.5*log10(sum(SFs*exp10(-0.4*Ps)))
         elif plog>0:
            vals = log10(sum(SFs*exp10(Ps)))
         elif plog<0:
            vals = -log10(sum(SFs*exp10(-Ps)))
         elif pinv:
            vals = 1/sum(SFs/Ps)
         else:
            vals = sum(SFs*Ps)
      elif lage:
         vals = self.interp(lage,Zu(start))
      else:
         return -99.999
      return vals

class BCDerivedProperty(BCProperty):
   def __init__(self,prop1,prop2,func,kx,ky,l=0,verb=0):
      self.prop1 = prop1
      self.prop2 = prop2
      self.func = func
      self.prop = "%s(%s,%s)" % (self.func,self.prop1,self.prop2)
      self.load(l)
      self.fit(kx,ky,verb=verb)
   def load(self,l):
      m = [22,32,42,52,62,72]
      z = [0.0001,0.0004,0.004,0.008,0.02,0.05]
      p1 = [load1prop(u,self.prop1) for u in m]
      p2 = [load1prop(u,self.prop2) for u in m]
      a = asarray([x[0] for x in p1])
      o1 = asarray([x[1] for x in p1])
      o2 = asarray([x[1] for x in p2])
      o = self.func(o1,o2)
      z = asarray([z[j]*ones(a[j].shape,float32) for j in range(len(z))])
      self.z = log10(z/0.02)
      self.a = a
      self.o = o
      if l: self.o = getlog10(self.o)

class FSPSMass(BCProperty):
   def __init__(self,tck):
       self.tck = tck
   def interp(self,lage,feh):
      if type(lage) == numpy.ndarray and type(feh) == numpy.ndarray and lage.shape == feh.shape:
        return bisplev2(lage,feh,self.tck)
      else:
        return bisplev([lage],[feh],self.tck)


class BCModel:
   def __init__(self,verb=1):
       if verb: print "Loading..."
       bff = bcdir+"/spectra/bisplrep.p"
       bf = open(bff,"r")
       P = cPickle.load(bf)
       bf.close()
       if verb: print "... done loading."
       self.crval1,self.cdelt1,self.crpix1,self.cd1_1,self.dcflag,self.z,self.t,self.kx,self.ky,self.tx,self.ty,self.B = P
       self.nx = len(self.B)
       self.wave = (arange(self.nx)+1-self.crpix1[0])*self.cdelt1[0]+self.crval1[0]
   def spectrum(self,age,Z):
       agemake = log10(age)
       zmake = log10(Z)
       model = power(10,[bisplev(agemake,zmake,[self.tx,self.ty,b,self.kx,self.ky]) for b in self.B])
       return model

class ised:
   def __init__(self,i):
      print "opening",i
      self.name = i
      if self.name[-3:] == ".gz": uopen = gzip.open
      else: uopen = open
      self.file = uopen(self.name,"r")
   def getline(self):
      self.line = split(self.file.readline())
   def getword(self):
      w = self.line[0]
      if len(self.line) > 1: self.line = self.line[1:]
      else: self.getline()
      return w
   def read(self):
      pos = 0
      self.getline()
      nsteps = int(self.getword())
      self.ages = asarray([float( self.getword()) for i in range(nsteps)])
      ml = long(self.getword())
      mu = long(self.getword())
      iseg  = long(self.getword())
      xx = []; lm = []; um = []; baux = []; cn = []; cc = []
      for i in range(iseg):
         xx.append( float(self.getword()))
         lm.append( float(self.getword()))
         um.append( float(self.getword()))
         baux.append( float(self.getword()))
         cn.append( float(self.getword()))
         cc.append( float(self.getword()))
      self.getline()
      id = join(self.line)
      self.getline()
      id2 = join(self.line)
      self.getline()
      id3 = join(self.line)
      self.getline()
      print id
      print id2
      print id3
      self.model = id
      inw = int(self.getword())
      self.wave = asarray([float( self.getword()) for i in range(inw)])
      self.spectra = []
      for n in range(nsteps):
         ik = int(self.getword())
         h = asarray([float( self.getword()) for i in range(ik)])
         self.spectra.append(h)
         ix = int(self.getword())
         f = asarray([float( self.getword()) for i in range(ix)])
         print n+1,;sys.stdout.flush()
      print
      for n in range(12):
         ik = int(self.getword())
         h = asarray([float( self.getword()) for i in range(ik)])
      self.spectra = asarray(self.spectra)
      self.file.close()

   def binread(self):
      self.data = self.file.read()
      self.file.close()
      hdat = fromstring(self.data[:12],int32)
      h1 = (nonzero(equal(hdat,0))[0]+1)*4
      l1 = (hdat[1]-1)*4
      self.ages = fromstring(self.data[h1-4:h1+l1],float32)
      self.data = self.data[h1+l1:]
      self.model = strip(self.data[24:92])
      print self.model
      nw = fromstring(self.data[292:296],int32)[0]
      self.wave = fromstring(self.data[296:296+nw*4],float32)
      ints = fromstring(self.data,int32)
      marks = nonzero(equal(ints,nw))+1
      self.spectra = []
      for m in marks:
         self.flux = fromstring(self.data[m*4:(m+nw)*4],float32)
         self.spectra.append(self.flux)
   def write(self,w1,w2,a1,a2):
#       self.spectra = self.spectra[1:]
       use = between(self.wave,w1,w2,inc=1)
       w = compress(use,self.wave,0)
       f = compress(use,self.spectra,1)
       tcks = [splrep(w,f1,k=3,s=0) for f1 in f]
       dw = mean(diff(w))
       w = arange(w1,w2+dw/2.,dw,float64)
       f = asarray([splev(w,t1) for t1 in tcks])
       use = between(self.ages,a1,a2,inc=1)
       f = compress(use,f,0)
       a = compress(use,self.ages,0)
       print log10(extrema(a)),a1,a2
       fitsout = replace(self.name,".ised_ASCII.gz",".fits")
       F = FITS(fitsout,create=1)
       F.newhead("T",-32,2,[w.shape[0],f.shape[0]])
       F.modkey("object",self.model)
       F.modkey("dc-flag",0)
       F.modkey("crpix1",1.0)
       F.modkey("cdelt1",mean(diff(w)))
       F.modkey("cd1_1",mean(diff(w)))
       F.modkey("crval1",w[0])
       [F.modkey("age"+str(r+1),a[r]) for r in range(f.shape[0])]
       [F.writerow(f[r],r) for r in range(f.shape[0])]
       F.close()
   def spline(self,kx=3,ky=3,a1=0,a2=1e24,w1=0,w2=1e24):
       tx = self.wave
       ty = self.ages
       x = ravel(len(self.ages)*[self.wave])
       y = ravel(self.ages[::,NewAxis]*ones(self.wave.shape)[NewAxis,::])
       z = self.spectra.ravel()
       keep = between(x,w1,w2,inc=1)*between(y,a1,a2,inc=1)
       x,y,z = compress(keep,[x,y,z],1)
       keep = between(tx,w1,w2,inc=1)
       tx = compress(keep,tx)
       keep = between(ty,a1,a2,inc=1)
       ty = compress(keep,ty)
       y = log10(y)
       ty = log10(ty)
       z = log10(z)
       self.tck = bisplrep(x,y,z,task=-1,s=None,kx=kx,ky=ky,tx=tx,ty=ty,nxest=len(tx),nyest=len(ty),full_output=0,quiet=0)
   def pickle(self,output):
       output = os.path.basename(output)
       print "Writing",output
       tx,ty,c,kx,ky = self.tck
       P = [self.wave,self.ages,self.tck]
       bf = gzip.open(output,"w")
       cPickle.dump(P,bf,pickle.HIGHEST_PROTOCOL)
       bf.close()
       print "Testing loading",output
       tf = gzip.open(output,"r")
       test = cPickle.load(tf)
       tf.close()
       print "Done loading",output

