
import cPickle
from scipy.special import *
from Expector import *
from VTKHelperFunctions import *
from masks import *
from scipy.interpolate import *
from scipy.optimize import leastsq
from quickhull2d import qhull

gl = True

argv,useldss3 = rdarg(argv,"-ldss3",None,0,single=1)
argv,useimacs= rdarg(argv,"-imacs",None,0,single=1)

if useldss3 or os.environ.get("LDSS3",False):
#   print "USING LDSS3!"
   useldss3 = True
   os.environ["LDSS3"] = "True"
elif useimacs or os.environ.get("IMACS",False):
#   print "USING IMACS!"
   useimacs = True
   os.environ["IMACS"] = "True"
elif not useldss3 and not useimacs:
#   print "DEFAULTING TO IMACS!"
   useimacs = True
   os.environ["IMACS"] = "True"

if useimacs:
   MM_ARCS = 1/2.880
   MM_MINS = 60*MM_ARCS
   ARCS_PIX = 0.2
   MINS_PIX = ARCS_PIX/60.0
   MAX_MINS = 15.0
   nccds = 8
elif useldss3:
   MM_ARCS = 1/2.880
   MM_MINS = 60*MM_ARCS
   ARCS_PIX = 0.188
   MINS_PIX = ARCS_PIX/60.0
   MAX_MINS = 8.0
   nccds = 1

def svdfit(b,y,th=1e-99):
      decomp = singular_value_decomposition(b)
      sol1 = transpose(decomp[2])
      useful = greater(decomp[1],th*decomp[1][0])
      sol2 = divz(1.0,useful*decomp[1])
      sol3 = dot(transpose(decomp[0]),y)
      if sometrue(sol3):
        solr = (sol2*sol3)
        soll = dot(sol1,solr)
      else:
        soll = zeros(sol3.shape)
      return soll


class SMF(DictType):
        SMFKEYS = ["type","id","ra","dec","x","y","w","t","b"]
        fn = [str,str,str,str,float,float,float,float,float]
        def ReadFile(self,smf_file):
            for key in self.SMFKEYS: self[key] = []
            f = open(smf_file,"r")
            for l in f:
               data = l.split()
               if (data[0] == "SLIT"):
                  [self[key].append(self.fn[k](data[k])) for k,key in enumerate(self.SMFKEYS)]
               elif (data[0] == "HOLE" and data[7] == "1"):
                  data.pop(7)
                  [self[key].append(self.fn[k](data[k])) for k,key in enumerate(self.SMFKEYS)]
            f.close()
            for key in self.SMFKEYS[4:]: self[key] = array(self[key])
        def GetObject(self,object):
            i = self["id"].index(object)
            vals = [self[key][i] for key in self.SMFKEYS]
            return vals


def NewCENs(targ):
    image, chip, xin, yin, fwhm, slitlen = targ
    return NewCEN( image, chip, xin, yin, fwhm, slitlen)

def NewCEN( image, chip, xin, yin, fwhm, slitlen, xzero=0, yzero=0):
    if isinstance(image,ArrayType):
       naxis2,naxis1 = image.shape
       x0,y0 = xin-xzero,yin-yzero
       xc = clip(floor([x0-2*fwhm,x0+3*fwhm+1]).astype(int32),0,naxis1-1)
       yc = clip(floor([y0-3*fwhm,y0+3*fwhm+1]).astype(int32),0,naxis2-1)
       d = median(transpose(image[yc[0]:yc[1],xc[0]:xc[1]]))
    elif isinstance(image,StringType):
       x0,y0 = xin-xzero,yin-yzero
       root = image.replace("b.fits","")[:-1]
       image = "%s%db.fits" % (root,chip)
       F = FITS(image)
       naxis1,naxis2 = F.naxes()
       xc = clip(floor([x0-2*fwhm,x0+3*fwhm+1]).astype(int32),0,naxis1-1)
       yc = clip(floor([y0-3*fwhm,y0+3*fwhm+1]).astype(int32),0,naxis2-1)
       d = median(transpose(F.readsec([xc[0],yc[0]],[xc[1],yc[1]])))
       F.close()
    a = yzero+yc[0]+arange(0.0,len(d),1.0,float32)
    p,w = centery(a,d,y0,fwhm=fwhm)
    dp = abs(p-yin)
    if dp < 0.5*fwhm: return p,w
    else: return yin,0.0

def centery(x,s,m,fwhm=5.0):
    t = array(edgy(x,s,m,0.85,fwhm))
    p = add.reduce(t)/2.0
    f = array(edgy(x,s,m,0.5,fwhm))
    h = abs(subtract.reduce(f))
    return p,h

def edgy(x,s,l,h=0.5,fwhm=5.0):
    a = int(max((x[0],l-5*fwhm)))
    b = int(min((x[-1],l+5*fwhm)))
    ux,us = compress(between(x,a,b,inc=1),[x,s],1)
    if len(ux): tck = splrep(ux,us,s=0)
    else: return 0.0,0.0
    v = splev(float(l),tck)
    vtck = splrep(ux,us-h*v,s=0)
    z = sproot(vtck,mest=len(ux))
    z = take(z,argsort(abs(z-l)))
    if len(z) > 1: return z[:2]
    else: return l,l


def invsoln(c,naxis1,naxis2,verb=0):
    nterms = len(c)
    x = asarray( range(0,naxis1+naxis1/nterms,naxis1/nterms) + [naxis1],float)
    y = asarray( range(0,naxis2+naxis2/nterms,naxis2/nterms) + [naxis2],float)
    x = VTKSubtract(x, naxis1/2.)
    y = VTKSubtract(y, naxis2/2.)
    nx = len(x); ny = len(y)
    o = ones((ny,nx),float)
    yy = (o*y[::,NewAxis]).ravel()
    xx = (o*x[NewAxis,::]).ravel()
    u,v = mbasis(c,xx,yy)
    nterm = [4,6,12,20,30,42,56]
    nords = [0,1, 3, 4, 5, 6, 6]
    m = nords[nterm.index(nterms)]
    b = abasis(m,u,v)
    ic = svdfit(b,concatenate([xx,yy]))
    xf,yf = mbasis(ic,u,v)
    xres = VTKSubtract(xf,xx); yres = VTKSubtract(yf,yy)
    if verb:
#      print "Using",len(c),"terms (order=",m,")"
      print "Inverse solution RMS=",rms(xres),rms(yres)
#      print "Inverse solution Extrema=",extrema(xres),extrema(yres)
    return ic


def VCP(x,v,minsides=0,tmp="/tmp/pd.dat",minarea=2.0,maxarea=100,verb=1,pad=0,maxlen=0):
    i = VTKMirrorPad(x,[-pad,x.shape[1]+pad,-pad,x.shape[0]+pad],numret=0)
    d = i.GetDimensions()
#    c = vtkMarchingContourFilter()
    c = vtkContourFilter()
#    c = vtkMarchingSquares()
    [c.SetValue(j,v[j]) for j in range(len(v))]
    c.ComputeScalarsOn()
#    c.ComputeNormalsOn()
#    c.ComputeGradientsOn()
#    c.UseScalarTreeOn()
#    p = c.GetOutput()
    c.SetInput(i)
#    c.Update()
#    p.GetBounds()
    pc = vtkPolyDataConnectivityFilter()
    pc.SetExtractionModeToAllRegions()
    pc.ScalarConnectivityOff()
    pc.ColorRegionsOn()
    pc.SetInput(c.GetOutput())
    pc.Update()
    N = pc.GetNumberOfExtractedRegions()
    regions = arange(N)

    d = []
    n = 1
    if verb: print "N=",N

    ps = vtkStripper()
    if maxlen: ps.SetMaximumLength(maxlen)
    else: ps.SetMaximumLength(x.shape[0]*x.shape[1])
    ps.SetInput(pc.GetOutput())
    po = ps.GetOutput()
    ps.Update()
    nl = po.GetNumberOfLines()
    nc = po.GetNumberOfCells()
    print "Number left=",nl
    areas = []
    for n in range(nl):
       pl = po.GetCell(n)
       x1,x2,y1,y2 = pl.GetBounds()[:4]
       a = (x2-x1)*(y2-y1)
       if a>minarea/3:
          pd = pl.GetPoints().GetData()
          pdx, pdy = VTKArrayToNum(pd)[:2]
          np = pl.GetNumberOfPoints()
          if np > 4:
             segments = asarray([[(pdx[j],pdx[j+1]),(pdy[j],pdy[j+1])] for j in range(len(pdx)-1)])
             if Closed(segments,verb=1):
                x1,y1 = transpose(asarray(segments)[::,::,0])
                x2,y2 = transpose(asarray(segments)[::,::,1])
                a = 0.5*abs(sum(x1*y2-y1*x2))
                if a >= minarea and a<=maxarea:
                   d.append(segments)
                   areas.append(a)
       if verb and n and not (n % (nl/80+1)):
          sys.stdout.write(".")
          sys.stdout.flush()

    if verb and N:
        print
        print "Left with",len(d)
        if d: print "Areas",extrema(areas),quartiles(array(areas))

    return d

def VEC(ix,contours,verb=0,speed=0,pixels=0,xzero=0,yzero=0,func=sum):
    enclosed = []
    xcens = []
    ycens = []
    mags = []
    ims = {}
    for jc,contour in enumerate(contours):
       if verb:
          if jc and not (jc % (len(contours)/40+1)):
             sys.stdout.write(".")
             sys.stdout.flush()
       (xb,yb),(xe,ye) = transpose(contour)
       x0,x1 = int(min(xb)), int(max(xb)+1.0)
       y0,y1 = int(min(yb)), int(max(yb)+1.0)
       xwid, ywid = x1-x0, y1-y0
       im = ix
       mcontour = 1.*contour
       mag = 1.0
       (xb,yb),(xe,ye) = transpose(mcontour)
       x0,x1 = int(min(xb)), int(max(xb)+1.0)
       y0,y1 = int(min(yb)), int(max(yb)+1.0)
       x0,x1 = clip([x0,x1],0,im.shape[1]-1)
       y0,y1 = clip([y0,y1],0,im.shape[0]-1)
       if y0>0: ixc = im[y0:y1].astype(ix.dtype)
       else: ixc = im[:y1].astype(ix.dtype)
       if x0>0: ixc = ixc[::,x0:x1].astype(ix.dtype)
       else: ixc = ixc[::,:x1].astype(ix.dtype)
       yco,xco = indices(ixc.shape)
       t = Interiors(xco.ravel()+x0,yco.ravel()+y0,mcontour)
       interior = less(abs(t-2*pi),1e-4)
       n = reshape(interior,ixc.shape)
       nixc = n*ixc
       yco,xco = indices(n.shape,float32)
       enc1 = func(nixc.ravel())
       wty = divz(sum((yco*nixc).ravel()),sum(nixc.ravel()))+y0+yzero
       wtx = divz(sum((xco*nixc).ravel()),sum(nixc.ravel()))+x0+xzero
       ycens.append(wty)
       xcens.append(wtx)
       enclosed.append(enc1)
    if verb: print
    xcens,ycens,enclosed = asarray([xcens,ycens,enclosed])
    return xcens,ycens,enclosed

def VDC(ix,contours,color=1,fill=0,verb=0):
#    ix = ix.astype(float64)
    if type(ix) == numpy.ndarray: i = NumToVTKImage(ix)
    else: i = ix
    c = vtkImageCanvasSource2D()
    e = i.GetExtent()
    c.SetExtent(e[0],e[1],e[2],e[3],e[4],e[5])
    c.Update()
#    c.SetScalarType( VTK_DOUBLE)
    c.SetScalarType( i.GetScalarType())
    c.DrawImage(0,0,i)
#    c.DrawImage(0,0,i,1,1,ix.shape[1],ix.shape[0])
    c.Update()
    for jc,contour in enumerate(contours):
       if verb:
          if jc and not (jc % (len(contours)/40+1)):
             sys.stdout.write(".")
             sys.stdout.flush()
       if isinstance(color,ArrayType) or isinstance(color,ListType): color1 = color[jc]
       else: color1 = color
       if isinstance(fill,ArrayType) or isinstance(fill,ListType): fill1 = fill[jc]
       else: fill1 = fill
       if fill1:
          (xb,yb),(xe,ye) = transpose(contour)
          x0,x1 = int(min(xb)), int(max(xb)+0.5)
          y0,y1 = int(min(yb)), int(max(yb)+0.5)
          cx1,cy1 = 1.*transpose(asarray(contour)[::,::,0]).astype(float64)
          cx2,cy2 = 1.*transpose(asarray(contour)[::,::,1]).astype(float64)
          if (cx1[-1],cy1[-1]) == (cx1[-1],cy1[-1]):
             cx1,cy1 = cx1[:-1],cy1[:-1]
          cx1 = clip(cx1,0,ix.shape[1]-1)
          cy1 = clip(cy1,0,ix.shape[0]-1)
          cz1 = zeros(cx1.shape,cx1.dtype).astype(float64)
          cp1 = 1.*asarray([cx1-x0,cy1-y0,cz1])
          ca1 = NumToVTKArray(cp1[0])
          ca2 = NumToVTKArray(cp1[1])
          ca3 = NumToVTKArray(cp1[2])
          ca = vtkDoubleArray()
          ca.SetNumberOfComponents(3)
          ca.SetNumberOfTuples(cp1.shape[1])
          ca.CopyComponent(0,ca1,0)
          ca.CopyComponent(1,ca2,0)
          ca.CopyComponent(2,ca3,0)
          po = vtkPoints()
          po.SetData(ca)
          cl = vtkPolyData()
          cl.SetPoints(po)
          cl.Update()
          sl = vtkImplicitSelectionLoop()
          sl.AutomaticNormalGenerationOn()
          sl.SetLoop(po)
          ii = vtkImplicitFunctionToImageStencil()
          ii.SetInput(sl)
          ii.SetOutputWholeExtent(0,x1-x0,0,y1-y0,e[4],e[5])
          t = vtkImageStencil()
          t.SetStencil( ii.GetOutput())
#          ixc = ix[y0:y1,x0:x1].astype(ix.dtype)
#          ic = NumToVTKImage(ixc)
          o2 = c.GetOutput()
          c.Update()
          o2 = vtkImageReslice()
          o2.SetInput( c.GetOutput())
          o2.SetOutputOrigin(x0,y0,e[4])
          o2.SetOutputExtent(0,x1-x0,0,y1-y0,0,e[5]-e[4])
          ic = o2.GetOutput()
          o2.Update()
          ic.SetOrigin(0,0,0)
          t.SetInput(ic)
          t.SetBackgroundValue(fill1)
          t.ReverseStencilOn()
          o = t.GetOutput()
          o.Update()
          c.DrawImage(x0,y0,o)
       for segment in contour:
          if color1:
             x,y = floor(segment+0.5).astype(int32)
             x = clip(x,e[0],e[1])
             y = clip(y,e[2],e[3])
             c.SetDrawColor(color1)
             c.DrawSegment3D(x[0],y[0],0.0,x[1],y[1],0.0)

    if verb: print
    o = c.GetOutput()
    c.Update()
    n = VTKImageToNum(o)
    return n

def CullContours(c,ct,fwhm,slitlen):
    print "Combine contours!"
    contours = []
    allcontours = [asarray(c1) for c1 in c]
    print len(c),
    sys.stdout.flush()
    lxs = array([min(ravel(c1[::,0,::])) for c1 in c])
    lys = array([min(ravel(c1[::,1,::])) for c1 in c])
    hxs = array([max(ravel(c1[::,0,::])) for c1 in c])
    hys = array([max(ravel(c1[::,1,::])) for c1 in c])
    xs = [c1[::,0,::] for c1 in c]
    ys = [c1[::,1,::] for c1 in c]
    x0s = [xs1[::,0] for xs1 in xs]
    x1s = [xs1[::,1] for xs1 in xs]
    y0s = [ys1[::,0] for ys1 in ys]
    y1s = [ys1[::,1] for ys1 in ys]
    ds = [sqrt((x11-x01)**2 + (y11-y01)**2) for x01,x11,y01,y11 in zip(x0s,x1s,y0s,y1s)]
    l = [sum(ds1) for ds1 in ds]
    xsp = [max(xs1[::,0])-min(xs1[::,0]) for xs1 in xs]
    ysp = [max(ys1[::,0])-min(ys1[::,0]) for ys1 in ys]
    l = (hxs-lxs)
    h = (hys-lys)
    print bwt(l),bwt(h),
    lc = [(xsp1,ysp1,l1,h1,t1,c1) for xsp1,ysp1,l1,h1,t1,c1 in zip(xsp,ysp,l,h,ct,c) \
              if slitlen/2<xsp1<2*slitlen and fwhm<ysp1<3*fwhm]
    xs = [lc1[0] for lc1 in lc]
    ys = [lc1[1] for lc1 in lc]
    l = [lc1[2] for lc1 in lc]
    h = [lc1[3] for lc1 in lc]
    ct = [lc1[4] for lc1 in lc]
    c = [lc1[5] for lc1 in lc]
    print len(c)
    if len(c):
       print extrema(xs),extrema(ys),extrema(l),extrema(h)
    contours.extend( [asarray(c1) for c1 in c])
    return allcontours,contours,ct

def CullObjects( xcs,ycs,zcs,xbs,xes,ybs,yes,fwhm,slitlen,plot=0):
    order = argsort(zcs)
    xcs,ycs,zcs,xbs,xes,ybs,yes = take([xcs,ycs,zcs,xbs,xes,ybs,yes],order,1)
    dx = abs(xcs[::,NewAxis] - xcs[NewAxis,::])
    dy = abs(ycs[::,NewAxis] - ycs[NewAxis,::])
    ds = sqrt(dx**2+dy**2)
    dz = greater(zcs[::,NewAxis],zcs[NewAxis,::])
    bad = array([sum(1.*between(dy[j],0,fwhm)*between(dx[j],0,slitlen/2.)*dz[j]) for j in range(len(xcs))])
    good = logical_not(bad)*greater(zcs,0)
    print len(bad),extrema(bad),len(nonzero(bad))
    if plot:
        [drawpt1(xcs[k],ycs[k],ps=17,ch=0.5,ci=2) for k in nonzero(bad)]
    newgood = zeros(good.shape,good.dtype)
    put(newgood,order,good)
    return newgood

def OLDL():
      if k in "L":
         M["t"] = []
         M["l"] = []
         lpxs = zeros(lus.shape,float)
         lpys = zeros(lus.shape,float)
         targs = []
         objs = Set(los)
         for obj in objs:
             jo = nonzero( equal(los,obj))
             olXs,olYs,olus,olvs,olucs,olvcs,olcs,olxs,olys,olws,olos = \
                take( [lXs,lYs,lus,lvs,lucs,lvcs,lcs,lxs,lys,lws,los],jo,1)
             olpx = zeros(olcs.shape,float32)
             olpy = zeros(olcs.shape,float32)
             for chip in sort(T.keys()):
                 jc = nonzero( equal(olcs,chip))
                 if len(jc):
                    ulu,ulv = take([olus,olvs],jc,1)
                    ulx,uly = mbasis(T[chip]["ico"],ulu/60.0*pi/180,ulv/60.0*pi/180)
                    ulx = ulx + T[chip]["xc"]
                    uly = uly + T[chip]["yc"]
                    put(olpx,jc,ulx)
                    put(olpy,jc,uly)
             put(lpxs,jo,olpx)
             put(lpys,jo,olpy)
         for f in range(len(lpxs)):
             targ1 = (image, lcs[f], lpxs[f], lpys[f], fwhm, slitlen)
             targs.append(targ1)
         for chip in sort(T.keys()):
             root = image.replace("b.fits","")[:-1]
             image = "%s%db.fits" % (root,chip)
             lcat = "%s%db.pos" % (root,chip)
             gcat = "%s%db.cen" % (root,chip)
             uc = nonzero(equal(lcs,chip))
             print lcat,len(uc)
             lf = open(lcat,"w")
             for j in uc:
                 print >>lf, los[j], lpxs[j], lpys[j]
             lf.close()
         ns = globals()
         if len(targs) > 1 and NP > 1:
            newCs = RunMinitasks(NewCENs,targs,ns,NP=NP,chunksize=1)
         else:
            newCs = map(NewCENs,targs)
         newcens,newfws = transpose(newCs)
         print quartiles(newfws),extrema(newfws)
         print "Bad ones:",sort(list(Set(compress(equal(newfws,0.0),lws))))
         print sum(equal(newfws,0.0)), sum(1.*equal(newfws,0.0))/len(newfws)
         newlus = zeros(lus.shape,float)
         newlvs = zeros(lvs.shape,float)
         medfws = median(compress(newfws,newfws))
         for chip in sort(T.keys()):
             root = image.replace("b.fits","")[:-1]
             image = "%s%db.fits" % (root,chip)
             lcat = "%s%db.pos" % (root,chip)
             gcat = "%s%db.cen" % (root,chip)
             lf = open(gcat,"w")
             jc = nonzero(equal(lcs,chip))
             if len(jc):
                uxs,uys,uws = take([lpxs,newcens,newfws],jc,1)
                newcenu,newcenv = multiply(mbasis(T[chip]["co"],uxs-T[chip]["xc"],uys-T[chip]["yc"]),180/pi*60)
                put(newlus,jc,newcenu)
                put(newlvs,jc,newcenv)
                for j in jc:
                    print >>lf, los[j], lpxs[j], newcens[j], newfws[j]
                for j2 in nonzero(equal(tcs,chip)):
                    du = abs(tus[j2]-newcenu)/MINS_PIX
                    dv = abs(tvs[j2]-newcenv)/MINS_PIX
                    k2 = argmin(dv)
                    if dv[k2] < fwhm/2.0 and medfws/2.<uws[k2]<1.5*medfws:
                       tys[j2] = uys[k2]
                       tvs[j2] = newcenv[k2]
                       tws[j2] = newfws[k2]
                    else:
                       tws[j2] = 0.0
             lf.close()
#         tks = greater(tws,0.0)
#         print sum(1-tks),1.*sum(1-tks)/len(tks)
#         tus,tvs,tcs,txs,tys,tzs,tws = compress(tks,[tus,tvs,tcs,txs,tys,tzs,tws],1)

def GetOne(*args):
    redux = 1
    xzero,yzero = 0,0
    if len(args) == 1: args = args[0]
    smf,Sroot,images,testobj,flatfield,lampfield,science,lamp,flat,obj,objid,lamc,dlam,s1,s2,adu,rn,ky,dkx,slitlen,naxis1,naxis2,wfm,fwhm,llist,refine,secondpass = args
    if llist and llist[0] != "/": llist = "../%s" % (llist)
    gl,wave1,wave2,Cord,Word,Sord,S,arots = ReadWFM(wfm)
    pck = "%s/%s.pck"%(Sroot,objid)
    pcks = [pck1 for pck1 in glob("%s/*.pck"%(Sroot)) if pck1 != pck]
    fSMF = SMF()
    fSMF.ReadFile(smf)
    objind = arange(len(fSMF["id"]))
    slitlen = (fSMF["b"][searchsorted(objind,obj)]+fSMF["t"][searchsorted(objind,obj)])/MM_MINS/MINS_PIX/s1
    hole = fSMF["type"][searchsorted(objind,obj)] == "HOLE"

    of = open(pck,"rb")
    olXs,olYs,olus,olvs,olucs,olvcs,olcs,olxs,olys,olws,olos, \
      olXbs,olYbs,olubs,olvbs,olxbs,olybs,\
      olXos,olYos,oluos,olvos,olxos,olyos,\
      olXts,olYts,oluts,olvts,olxts,olyts, = cPickle.load(of)
    of.close()
    chips = sort(list(Set(olcs)))
    image = images[0]
    root = image.replace("b.fits","")[:-1]
    root2 = root[:-1]+"s"
    print objid,root,chips
    nns = False
    for chip in chips[::-1]:
        uc = equal(olcs,chip)
        clXs,clYs,clus,clvs,clucs,clvcs,clcs,clxs,clys,clws,clos = compress(uc, \
           [olXs,olYs,olus,olvs,olucs,olvcs,olcs,olxs,olys,olws,olos],1)
        clXbs,clYbs,clubs,clvbs,clxbs,clybs = compress(uc,[olXbs,olYbs,olubs,olvbs,olxbs,olybs],1)
        clXos,clYos,cluos,clvos,clxos,clyos = compress(uc,[olXos,olYos,oluos,olvos,olxos,olyos],1)
        clXts,clYts,cluts,clvts,clxts,clyts = compress(uc,[olXts,olYts,oluts,olvts,olxts,olyts],1)

#        culx,culy = mbasis(T[chip]["ico"],clucs/60.0*pi/180,clvcs/60.0*pi/180)
#        culx = culx + T[chip]["xc"] - 1.0
#        culy = culy + T[chip]["yc"] - 1.0

        if chip in [5,6,7,8]:
           order = argsort(clys)[::-1]
           clXs,clYs,clus,clvs,clucs,clvcs,clcs,clxs,clys,clws,clos = \
                   take([clXs,clYs,clus,clvs,clucs,clvcs,clcs,clxs,clys,clws,clos],order,1)
           clXbs,clYbs,clubs,clvbs,clxbs,clybs = \
                   take([clXbs,clYbs,clubs,clvbs,clxbs,clybs],order,1)
           clXos,clYos,cluos,clvos,clxos,clyos = \
                   take([clXos,clYos,cluos,clvos,clxos,clyos],order,1)
           clXts,clYts,cluts,clvts,clxts,clyts = \
                   take([clXts,clYts,cluts,clvts,clxts,clyts],order,1)

        bimage = "%s%db.fits" % (root,chip)
        himage = "%s%dbh.fits" % (root,chip)

        ymin,ymax = clip(floor(extrema(clys-yzero)).astype(int32),0,naxis2-1)

        cchips = {1:6,6:1,2:5,5:2,3:8,8:3,4:7,7:4}

        if cchips[chip] in chips:
           ymax = naxis2-1

        xbl,xbh = extrema(clxbs)-xzero
        xtl,xth = extrema(clxts)-xzero
        xc = clip(floor(extrema([xbl,xbh,xtl,xth])).astype(int32),0,naxis1-1)
        xc = clip(add(xc,[-5,5]),0,naxis1-1)
        yc = [ymin,ymax]
        yr = 0.5*(ymax+ymin)+0

        print xc[1]-xc[0],slitlen

        if (ymax-ymin)>20 and (xc[1]-xc[0])>=0.97*slitlen:

            print os.path.basename(pck),
            fmt = 4*" %5.0f"
            print fmt % (yc[0], yc[1], xc[0], xc[1])

            if os.path.exists("good.fits"):
               F = FITS("good.fits")
               goodpix = F.readsec([xc[0],yc[0]],[xc[1],yc[1]])
               F.close()
            else:
               goodpix = 1

            F = FITS(bimage)
            naxis1,naxis2 = F.naxes()
            print bimage,naxis1,xc,naxis2,yc
            d = F.readsec([xc[0],yc[0]],[xc[1],yc[1]])
            d = goodpix*d
            dweltime = F["dweltime"]
            ncycles = F["ncycles"]
            nshuffle = F["nshuffle"]
            if F["cpdshift"] != "N/A":
               cpdshift = "-s %f" % (F["cpdshift"])
            else:
               cpdshift = "-fft"
            if dweltime and ncycles and nshuffle:
               nns = True
               d2 = F.readsec([xc[0]+nshuffle,yc[0]],[xc[1]+nshuffle,yc[1]])
               d2 = d2[:d.shape[0],:d.shape[1]]
               d2 = d2*goodpix
            F.close()
            d = transpose(d)
            if nns:
               d2 = transpose(d2)
            if flat:
               F = FITS(bimage.replace(".fits","z.fits"))
               dz = F.readsec([xc[0],yc[0]],[xc[1],yc[1]])
               F.close()
               if useldss3:
                  Fstep = FITS(himage)
                  step = Fstep.data()
                  Fstep.close()
                  print "Dividing by step!"
                  d = divz(d,step[yc[0]:yc[1]+1][NewAxis,::])
               dz = transpose(dz)
               d = d*equal(dz,0)

            sxi = (d.shape[0]-1.0)
            syi = (d.shape[1]-1.0)

            xi0,yi0 = indices(d.shape,int32)
            xi0 = xi0 + xc[0]
            yi0 = yi0 + yc[0]

            if chip in [5,6,7,8]:
               d = d[::-1,::-1]
               if nns:
                  d2 = d2[::-1,::-1]
               ymin,ymax = (naxis2-(1-yzero))-ymax,(naxis2-(1-yzero))-ymin
               yr = 0.5*(ymax+ymin)-0
               clys = (naxis2-(1-yzero))-clys
               clybs = (naxis2-(1-yzero))-clybs
               clyos = (naxis2-(1-yzero))-clyos
               clyts = (naxis2-(1-yzero))-clyts

            if testobj:
               qdump("/tmp/d%s.fits"%(objid),d)
               raw_input("<hit cr> ")
            povers = zeros(d.shape,float32)
            if (flat and not hole) or testobj:
               print "Checking for overlaps!"
               for pck1 in pcks:
                  pf = open(pck1,"rb")
                  plXs,plYs,plus,plvs,plucs,plvcs,plcs,plxs,plys,plws,plos, \
                    plXbs,plYbs,plubs,plvbs,plxbs,plybs,\
                    plXos,plYos,pluos,plvos,plxos,plyos,\
                    plXts,plYts,pluts,plvts,plxts,plyts, = cPickle.load(pf)
                  pf.close()
                  pchips = sort(list(Set(plcs)))
                  puc = equal(plcs,chip)
                  pobjid = fSMF["id"][searchsorted(objind,plos[0])]
                  pslitlen = (fSMF["b"][searchsorted(objind,plos[0])]+fSMF["t"][searchsorted(objind,plos[0])])/MM_MINS/MINS_PIX/s1
                  if puc.any():
                     pclXs,pclYs,pclus,pclvs,pclucs,pclvcs,pclcs,pclxs,pclys,pclws,pclos = compress(puc, \
                        [plXs,plYs,plus,plvs,plucs,plvcs,plcs,plxs,plys,plws,plos],1)
                     pclXbs,pclYbs,pclubs,pclvbs,pclxbs,pclybs = compress(puc,[plXbs,plYbs,plubs,plvbs,plxbs,plybs],1)
                     pclXos,pclYos,pcluos,pclvos,pclxos,pclyos = compress(puc,[plXos,plYos,pluos,plvos,plxos,plyos],1)
                     pclXts,pclYts,pcluts,pclvts,pclxts,pclyts = compress(puc,[plXts,plYts,pluts,plvts,plxts,plyts],1)

                     if chip in [5,6,7,8]:
                        order = argsort(pclys)[::-1]
                        pclXs,pclYs,pclus,pclvs,pclucs,pclvcs,pclcs,pclxs,pclys,pclws,pclos = \
                                take([pclXs,pclYs,pclus,pclvs,pclucs,pclvcs,pclcs,pclxs,pclys,pclws,pclos],order,1)
                        pclXbs,pclYbs,pclubs,pclvbs,pclxbs,pclybs = \
                                take([pclXbs,pclYbs,pclubs,pclvbs,pclxbs,pclybs],order,1)
                        pclXos,pclYos,pcluos,pclvos,pclxos,pclyos = \
                                take([pclXos,pclYos,pcluos,pclvos,pclxos,pclyos],order,1)
                        pclXts,pclYts,pcluts,pclvts,pclxts,pclyts = \
                                take([pclXts,pclYts,pcluts,pclvts,pclxts,pclyts],order,1)

                     pymin,pymax = clip(floor(extrema(pclys-1)).astype(int32),0,naxis2-1)

                     if cchips[chip] in pchips:
                         pymax = naxis2-1

                     pxbl,pxbh = extrema(pclxbs)-xzero
                     pxtl,pxth = extrema(pclxts)-xzero
                     pxc = clip(floor(extrema([pxbl,pxbh+1,pxtl,pxth+1])).astype(int32),0,naxis1-1)
                     pyc = [pymin,pymax]
                     pyr = 0.5*(pymax+pymin)+0
                     print os.path.basename(pck1),
                     fmt = 4*" %5.0f"
                     print fmt % (pyc[0], pyc[1], pxc[0], pxc[1])

                     if ((yc[0] <= pyc[0] <= yc[1]) or (yc[0] <= pyc[1] <= yc[1])) and \
                             ((xc[0] <= pxc[0] <= xc[1]) or (xc[0] <= pxc[1] <= xc[1])):

                        F = FITS(bimage)
                        pd = F.readsec([pxc[0],pyc[0]],[pxc[1],pyc[1]])
                        F.close()
                        pd = transpose(pd)

                        pxi0,pyi0 = indices(pd.shape,int32)
                        pxi0 = pxi0 + pxc[0]
                        pyi0 = pyi0 + pyc[0]
#                        if chip in [5,6,7,8]:
#                           pxi0 = (pd.shape[0]-1) - pxi0 + pxc[1]
#                           pyi0 = (pd.shape[1]-1) - pyi0 + pyc[1]
#                        else:
#                           pxi0 = pxi0 + pxc[0]
#                           pyi0 = pyi0 + pyc[0]

                        if chip in [5,6,7,8]:
                           pd = pd[::-1,::-1]
#                           pxi0 = pxi0[::-1,::-1]
#                           pyi0 = pyi0[::-1,::-1]
                           pymin,pymax = (naxis2-(1-yzero))-pymax,(naxis2-(1-yzero))-pymin
                           pclys = (naxis2-(1-yzero))-pclys
                           pclybs = (naxis2-(1-yzero))-pclybs
                           pclyos = (naxis2-(1-yzero))-pclyos
                           pclyts = (naxis2-(1-yzero))-pclyts
#                           pxi0 = naxis1-pxi0
#                           pyi0 = naxis2-pyi0

#                        qdump("/tmp/pd%s.fits"%(pobjid),pd)

                        psxi = pd.shape[0]/2.0
                        psyi = pd.shape[1]/2.0

                        pyr = 0.5*(pymax+pymin)-0
                        pnyi = divz(pclys-pyr,psyi)
                        uCord = min([Cord,len(pclxs)-1])
                        uCord = min([uCord,int((pymax-pymin)/300)])
                        pdxb = legendre(pnyi,uCord)
                        pdxs = svdfit(transpose(pdxb),pclxs)
                        pdxf = dot(pdxs,pdxb)
                        psr = rms(pdxf-pclxs)
                        pxr = pdxs[0]+1
                        pdxs = svdfit(transpose(pdxb),(pclxs-pxr)/pd.shape[0])
                        pdxf = dot(pdxs,pdxb)

                        pcsect1a = int(pxr-pxc[0]-pslitlen/2.)+1
                        pcsect1b = int(pxr-pxc[0]+pslitlen/2.+1)+1
                        if chip in [5,6,7,8]:
                           pdxs = -pdxs
                           pcsect1a,pcsect1b = pd.shape[0]-(pcsect1b-1)+0, pd.shape[0]-(pcsect1a-1)+0
                        pcsect1a = max([1,pcsect1a])
                        pcsect1b = min([pcsect1b,pd.shape[0]])
#                        print pcsect1a, pcsect1b,pslitlen
                        pmud = getmapping(0,uCord,pdxs,pd,ydist=1,verb=0).astype(float32)
                        pxco = reversebin(pxi0,0.0*pmud,pmud,data=0,cubic=0,constant=-999)
                        pd = 1.*between(pxi0-pxc[0]+1-0.5*(pcsect1a+pcsect1b),-pslitlen/2.,pslitlen/2.,inc=0)
                        pd = 1.*greater(reversebin(pd,0.0*pmud,pmud,data=0,cubic=0,constant=-999),0.5)
                        pover = pd
#                        print pobjid,pslitlen,slitlen
                        if testobj: qdump("/tmp/py%s.fits"%(pobjid),9999.*pd)

#                        qdump("/tmp/pxi0%s.fits"%(pobjid),pxi0)
#                        qdump("/tmp/pyi0%s.fits"%(pobjid),pyi0)

                        pset = {}
                        dpx0, dpy0 = (xc[0]-pxc[0]), (yc[0]-pyc[0])
                        dpx1, dpy1 = (xc[1]-pxc[1]), (yc[1]-pyc[1])
                        dpx2, dpy2 = 0.0, 0.0
                        if chip in [5,6,7,8]:
                            dpx2, dpy2 = dpx1+dpx0, dpy1+dpy0
#                        else:
#                            dpx2, dpy2 = 0.0, 0.0
#                        print extrema(pxi0.ravel()), extrema(pyi0.ravel())
#                        print dpx0,dpy0
#                        print dpx1,dpy1
#                        print dpx2,dpy2
                        [pset.setdefault((x1+dpx2,y1+dpy2),z1) for x1,y1,z1 in zip(pxi0.ravel(),pyi0.ravel(),pd.ravel())]
                        pkx,pky = transpose(pset.keys())
#                        print sort(pkx)
#                        print sort(pky)
                        pover = reshape([pset.get((x1,y1),0) for x1,y1 in zip(xi0.ravel(),yi0.ravel())],d.shape)
#                        print sort(list(Set(xi0.ravel())))
#                        print sort(list(Set(yi0.ravel())))
                        povers = povers + pover
#                        print sum(pover.ravel())
#                        pover = 9999.*pover
#
                        if testobj and (1 or pover.ravel().any()): qdump("/tmp/pd%s_over.fits"%(pobjid),9999.*pover)

            eimage = "%s/%s_%sc%d.fits"%(Sroot,root,objid,chip)
            if nns:
               eimage2 = "%s/%s_%sc%d.fits"%(Sroot,root2,objid,chip)
            eimagemud = "%s/%s_%sc%d-y-dist.fits"%(Sroot,root,objid,chip)
            eimagexmud = "%s/%s_%sc%d-x-rect.fits"%(Sroot,root,objid,chip)

            print obj,objid,chip,xc[0],xc[1],ymin,ymax,eimage
            if testobj: exit()

            nyi = divz(clys-yr,(d.shape[1]-1.0)/2.0)
            yspan = (max(clws)-min(clws))/dlam
            uCord = min([Cord,len(clxs)-1])
#            uCord = min([uCord,int((ymax-ymin)/300)])
#            uCord = min([uCord,int(yspan/300)])
            dxb = legendre(nyi,uCord)
            dxs = svdfit(transpose(dxb),clxs)
            dxf = dot(dxs,dxb)
            sr = rms(dxf-clxs)
            xr = dxs[0]+1
            dxs = svdfit(transpose(dxb),(clxs-xr)/(d.shape[0]-1.0))
            dxf = dot(dxs,dxb)

            csect1a = int(xr-xc[0]-slitlen/2.+0.5)+1
            csect1b = int(xr-xc[0]+slitlen/2.+0.5+1)+1

            if chip in [5,6,7,8]:
               objpos = (xc[1] - (clxos[len(clxos)/2]+1)) + 0
               objpos = d.shape[0] - ((clxos[len(clxos)/2]+1 - xc[0]) + 0)
               dxs = -dxs
               csect1a,csect1b = d.shape[0]-(csect1b-1)+1, d.shape[0]-(csect1a-1)+1
            else:
               objpos = (clxos[len(clxos)/2]+1 - xc[0]) + 1
            print chip,objpos

            csect1a = max([1,csect1a])
            csect1b = min([csect1b,d.shape[0]])

            uWord = min([Word,len(clws)-1])
            iwb = polynom(clws/1e4,uWord)
            iws = svdfit(transpose(iwb),nyi)
            dwb = legendre(nyi,uWord)
            dws = svdfit(transpose(dwb),clws)
            if chip in S:
               [wxsol,wxrms,wxbwt,wysol,wyrms,wybwt] = S[chip]
            else:
               [wxsol,wxrms,wxbwt,wysol,wyrms,wybwt] = S[0]
            wr = wybwt[1]*dws[1]/syi

            dys = (clys-yr)/syi
            dyts = (clyts-clys)/syi
            dycs = (clys-clys)/syi
            dyos = (clyos-clys)/syi
            dybs = (clybs-clys)/syi
            xx = concatenate([dys,dys,dys])
            yy = concatenate([-ones(dyts.shape),zeros(dyts.shape),ones(dyts.shape)])
            xord,yord = 1,2
            pb = prepbasis(xx,yy,xord,yord)
            xsol = svdfit(transpose(pb), concatenate([dyts,dycs,dybs]))
#            print clyts-clys
#            print clybs-clys
#            print xsol

            extras = [("nslits",1),("csect1a",csect1a),("csect1b",csect1b),("objpos",objpos)]
            dxsr = 1*dxs

            mud = getmapping(0,uCord,dxsr,d,ydist=1,verb=0).astype(float32)
            xmud = getmapping(xord,yord,xsol,d,ydist=0,verb=0).astype(float32)
            d = d*equal(povers,0)
            if nns:
               d2 = d2*equal(povers,0)

            if secondpass:
               F =FITS(eimage)
               objpos = F["objpos"]
               F.close()
            else:
               qdump(eimage,d,t=bimage,extras=extras,verb=0)
               if nns:
                  qdump(eimage2,d2,t=eimage,extras=extras,verb=0)
               qdump(eimagemud,mud,t=bimage,extras=extras,verb=0)
               qdump(eimagexmud,xmud,t=bimage,extras=extras,verb=0)

               efs = [eimage,eimagemud,eimagexmud]
               if nns: efs.append(eimage2)
               for ef in efs:
                   F=FITS(ef,readonly=0)
                   writesol(1,0,[dxsr,sr],F,keyroot="y0xfm",xord=0,yord=uCord,verb=0)
                   writesol(1,0,[xsol,sr],F,keyroot="x0xfm",xord=xord,yord=yord,verb=0)
                   F.modkey("xdcor",0)
                   F.modkey("ydcor",0)
                   F.modkey("y-dist",os.path.basename(eimagemud))
                   F.modkey("x-rect",os.path.basename(eimagexmud))
                   writesol(1,0,[dws,wr],F,keyroot="w0xfm",xord=uWord,yord=0,verb=0)
                   F.close()

#            exit()

#            qdump("/tmp/mud.fits",mud)
#            qdump("/tmp/xmud.fits",xmud)
#            os._exit(1)

            if redux:

               os.chdir(Sroot)
               eimage = os.path.basename(eimage)
               if nns:
                  eimage2 = os.path.basename(eimage2)

               logfile = "%s_%sc%d.log"%(root,objid,chip)
               if os.path.exists(logfile) and not secondpass:
                  os.remove("%s"%(logfile))
                  fl=open("%s"%(logfile),"w");fl.close()

               if flatfield:
                  fimage1d = "%sc_%sc%dslt.fits"%(flatfield,objid,chip)
                  fimage = "%sc_%sc%do.fits"%(flatfield,objid,chip)

                  if not secondpass:
                      ftest = "flat1d -ydist %s %s %s >> %s"%(cpdshift,fimage1d,eimage,logfile)
                      verbex("echo "+ftest)
                      verbex(ftest)
                      eimage = "%s_%sc%ds.fits"%(root,objid,chip)
                      if nns:
                         ftest = "flat1d -ydist %s %s %s >> %s"%(cpdshift,fimage1d,eimage2,logfile)
                         verbex("echo "+ftest)
                         verbex(ftest)
                         eimage2 = "%s_%sc%ds.fits"%(root2,objid,chip)

                      ftest = "flat2d -low 0.1 %s %s >> %s"%(fimage,eimage,logfile)
                      verbex("echo "+ftest)
                      verbex(ftest)
                      if nns:
                         ftest = "flat2d -low 0.1 %s %s >> %s"%(fimage,eimage2,logfile)
                         verbex("echo "+ftest)
                         verbex(ftest)
                  eimage = "%s_%sc%dsf.fits"%(root,objid,chip)
                  if nns:
                     eimage2 = "%s_%sc%dsf.fits"%(root2,objid,chip)

               mimage = os.path.basename(eimage).replace(".fits","m.fits")
               oimage = os.path.basename(eimage).replace(".fits","o.fits")
               srimage = os.path.basename(eimage).replace(".fits","r.fits")
               mrimage = os.path.basename(eimage).replace(".fits","mr.fits")
               orimage = os.path.basename(eimage).replace(".fits","or.fits")
               s1image = os.path.basename(eimage).replace(".fits","r01.fits")
               m1image = os.path.basename(eimage).replace(".fits","mr01.fits")
               o1image = os.path.basename(eimage).replace(".fits","or01.fits")
               if nns:
                  mimage2 = os.path.basename(eimage2).replace(".fits","m.fits")
                  oimage2 = os.path.basename(eimage2).replace(".fits","o.fits")
                  srimage2 = os.path.basename(eimage2).replace(".fits","r.fits")
                  mrimage2 = os.path.basename(eimage2).replace(".fits","mr.fits")
                  orimage2 = os.path.basename(eimage2).replace(".fits","or.fits")
                  s1image2 = os.path.basename(eimage2).replace(".fits","r01.fits")
                  m1image2 = os.path.basename(eimage2).replace(".fits","mr01.fits")
                  o1image2 = os.path.basename(eimage2).replace(".fits","or01.fits")

               if (not secondpass and lamp and llist) and (refine or useldss3):
                  llo = open(llist,"r")
                  lams = array([float(line.split()[0]) for line in llo if line and line[0] != "#"])
                  llo.close()
                  ulams = compress(between(lams,min(olws)+5*s2*dlam,max(olws)-5*s2*dlam),lams)
                  if len(ulams) > 2:
                     iwb = polynom(ulams/1e4,uWord)
                     ylams = syi*dot(iws,iwb)+yr
                     yspan = max(ylams)-min(ylams)
                     if useldss3 and not hole:
                        nx2 = min([int(yspan/300)+1,8])
                        if nx2>2: nsub2 = 2
                        else: nsub2 = 1
                        xord2 = min([int(yspan/300),2])
                        yord2 = 3
                        if slitlen > 20: yord2 = 4

                        gtest = "getrect -ydist -xdist -b 2 -dy 3 -nx %d -nsub2 %d -x %d -y %d" % (nx2,nsub2,xord2,yord2)
                        verbex("echo %s %s >> %s" % (gtest,eimage,logfile))
                        verbex("%s %s >> %s" % (gtest,eimage,logfile))

               if not secondpass and (science or lamp) and llist:
                  llo = open(llist,"r")
                  lams = array([float(line.split()[0]) for line in llo if line and line[0] != "#"])
                  llo.close()
                  ulams = compress(between(lams,min(clws)+5*s2*dlam,max(clws)-5*s2*dlam),lams)
                  if len(ulams) > 1:
                     iwb = polynom(ulams/1e4,uWord)
                     ylams = syi*dot(iws,iwb)+yr
                     yspan = max(ylams)-min(ylams)

                     if (useldss3 or refine) and not hole:
                        if lamp:
#                           Wzero = min([(uWord+1)/2,len(ulams)-1])
                           Wzero = min([uWord+1,(len(ulams))])
                           Wzero = min([int(yspan/200)+1,Wzero])
                           Wzero = -Wzero 
                        elif science: Wzero = 1
                        wtest = "waverect -ydist -xdist -wdist -g %f -rn %f -b 2 -bth 5 -arc -wide -cut 5 -fwhm %.1f -o %d -zero %d -ll %s" % (adu,rn,1.5*fwhm,uWord,Wzero,llist)
                        verbex("echo ulams %d %f %f >> %s" % (len(ulams),ulams[0],ulams[-1],logfile))
                        verbex("echo %s %s >> %s" % (wtest,eimage,logfile))
                        verbex("%s %s >> %s" % (wtest,eimage,logfile))
                        if nns:
                           wtest = "copyrect -wdist"
                           verbex("echo %s %s %s >> %s" % (gtest,eimage,eimage2,logfile))
                           verbex("%s %s %s >> %s" % (gtest,eimage,eimage2,logfile))

               if not secondpass and (lampfield and science):

                     if flatfield:
                        limage = "%sc_%sc%dsf.fits"%(lampfield,objid,chip)
                     else:
                        limage = "%sc_%sc%d.fits"%(lampfield,objid,chip)
                     gtest = "copyrect -xdist %s" % (limage)
                     verbex("echo %s %s >> %s" % (gtest,eimage,logfile))
                     verbex("%s %s >> %s" % (gtest,eimage,logfile))
                     gtest = "copyrect -wdist %s" % (limage)
                     verbex("echo %s %s >> %s" % (gtest,eimage,logfile))
                     verbex("%s %s >> %s" % (gtest,eimage,logfile))
                     if nns:
                        gtest = "copyrect -xdist %s" % (limage)
                        verbex("echo %s %s >> %s" % (gtest,eimage2,logfile))
                        verbex("%s %s >> %s" % (gtest,eimage2,logfile))
                        gtest = "copyrect -wdist %s" % (limage)
                        verbex("echo %s %s >> %s" % (gtest,eimage2,logfile))
                        verbex("%s %s >> %s" % (gtest,eimage2,logfile))

               if flat:
                  ky=0
                  if flat == 2: dkx = 0.7
                  else: dkx=21
                  skycut = 20.0
               elif science or lamp:
                  ky=ky
                  dkx=dkx
                  if lamp: skycut = 20.0
                  else: skycut = 10.0

               if nns:
                  stest = "cp %s %s" % (eimage2,mimage)
                  verbex("%s >> %s" % (stest,logfile))
                  stest = "cp %s %s" % (eimage,mimage2)
                  verbex("%s >> %s" % (stest,logfile))
                  stest = "addkeys %s skyim %s" % (eimage,mimage)
                  verbex("%s >> %s" % (stest,logfile))
                  stest = "addkeys %s skyim %s" % (eimage2,mimage2)
                  verbex("%s >> %s" % (stest,logfile))
               else:
                  if flat:
                     stest = "makeflat -star 0 -ydist -xdist"
                     verbex("%s %s >> %s" % (stest,eimage,logfile))
                     eimage1d = eimage.replace(".fits","slt.fits")
                     eimage2d = eimage.replace(".fits","flt.fits")
                     eimagesc = eimage.replace(".fits","s.fits")
                     eimagesfc = eimage.replace(".fits","sf.fits")

                     stest = "flat1d -ydist"
                     verbex("%s %s %s >> %s" % (stest,eimage1d,eimage,logfile))

                     stest = "flat2d -low 0.1"
                     verbex("%s %s %s >> %s" % (stest,eimage2d,eimagesc,logfile))

                     verbex("ln -fs %s %s >> %s" % (eimagesfc,mimage,logfile))
                     verbex("ln -fs %s %s >> %s" % (eimagesfc,mimage.replace("m.fits","sm.fits"),logfile))
                     verbex("ln -fs %s %s >> %s" % (eimage2d,oimage,logfile))

                  else:
                     stest = "skyrect -new -ydist -xdist -b 2 -skyint 4096 -kx 3 -dkx %f -ky %d -g %f -rn %f -spass 3 -skycut %f -check %d -pct 45" % (dkx,ky,adu,rn,skycut,int(slitlen/2))
#                     skyaps = "%s_%sc%df.sky"%(root,objid,chip)
                     skyaps = eimage.replace(".fits",".sky")
                     if os.path.exists(skyaps):
                        stest = stest + " -skysuf sky"
                     verbex("%s %s >> %s" % (stest,eimage,logfile))

#               if flat:
#                  print useldss3,useimacs
#                  if useldss3:
#                     otest = "efits %s %s ../%s \"i3[NewAxis,%d:%d]*divz(i1,i2)*not_equal(i1,0)\" %s >> %s"
#                     verbex(otest%(eimage,mimage,himage,yc[0],yc[1]+1,oimage,logfile))
#                  elif useimacs:
##                     otest = "efits -r %s %s \"tmp=divz(i1,i2)*not_equal(i1,0);result=divz(tmp,VTKMedian(tmp,21,1,numret=1))\" %s >> %s"
#                     otest = "efits %s %s \"divz(i1,i2)*not_equal(i1,0)\" %s >> %s"
#                     verbex(otest%(eimage,mimage,oimage,logfile))
               if science or lamp:
                  if nns:
                     otest = "efits %s %s %s \"(i1-i2)*not_equal(i1,0)*less(i2,i3+5*getsqrt(i3/%f+(%f/%f)**2))\" %s >> %s"
                     verbex(otest%(eimage,mimage,mimage2,adu,rn,adu,oimage,logfile))
                     otest = "efits %s %s %s \"(i1-i2)*not_equal(i1,0)*less(i2,i3+5*getsqrt(i3/%f+(%f/%f)**2))\" %s >> %s"
                     verbex(otest%(eimage2,mimage2,mimage,adu,rn,adu,oimage2,logfile))
                  else:
                     otest = "efits %s %s \"(i1-i2)*not_equal(i1,0)\" %s >> %s"
                     verbex(otest%(eimage,mimage,oimage,logfile))

#               os._exit(1)

########        ttest = "unrect -ydist >> %s"
########        verbex("%s %s" % (ttest,eimage,logfile))
########        verbex("%s %s" % (ttest,mimage,logfile))
########        verbex("%s %s" % (ttest,oimage,logfile))

               if os.path.exists(eimage) and os.path.exists(mimage) and os.path.exists(oimage):
                   if 0 and lamp:
                      ctest = "conrect -image %s -sky %s -obj %s -slit 1 -p 5" % (eimage,mimage,oimage)
                      verbex("%s >> %s" % (ctest,logfile))
                      mimage = os.path.basename(eimage).replace(".fits","m2.fits")
                      oimage = os.path.basename(eimage).replace(".fits","o2.fits")
                      mrimage = os.path.basename(eimage).replace(".fits","m2r.fits")
                      orimage = os.path.basename(eimage).replace(".fits","o2r.fits")
                      m1image = os.path.basename(eimage).replace(".fits","m2r01.fits")
                      o1image = os.path.basename(eimage).replace(".fits","o2r01.fits")


                   rtest = "unrect -ydist -xdist -wdist -w1 %f -w2 %f -dw %f" % (wave1,wave2,s2*dlam)
########            rtest = "unrect -ydist -xdist"
                   verbex("echo %s %s >> %s" % (rtest,eimage,logfile))
                   verbex("%s %s >> %s" % (rtest,eimage,logfile))
                   verbex("echo %s %s >> %s" % (rtest,mimage,logfile))
                   verbex("%s %s >> %s" % (rtest,mimage,logfile))
                   if science or lamp:
                      verbex("echo %s %s >> %s" % (rtest,oimage,logfile))
                      verbex("%s %s >> %s" % (rtest,oimage,logfile))
                   if nns:
                      verbex("echo %s %s >> %s" % (rtest,eimage2,logfile))
                      verbex("%s %s >> %s" % (rtest,eimage2,logfile))
                      verbex("echo %s %s >> %s" % (rtest,mimage2,logfile))
                      verbex("%s %s >> %s" % (rtest,mimage2,logfile))
                      if science or lamp:
                         verbex("echo %s %s >> %s" % (rtest,oimage2,logfile))
                         verbex("%s %s >> %s" % (rtest,oimage2,logfile))

                   etest = "efits %s \"i1[%d:%d]\" %s >> %s"
                   htest = "addkeys %s csect1a,csect1b 1,"+str(csect1b-csect1a+1)+" -i >> %s"
                   jtest = "addkeys %s objpos "+str(objpos-csect1a+1)+" -f >> %s"
                   verbex(etest % (srimage,csect1a-1,csect1b,s1image,logfile))
                   verbex(htest % (s1image,logfile))
                   verbex(jtest % (s1image,logfile))
                   verbex(etest % (mrimage,csect1a-1,csect1b,m1image,logfile))
                   verbex(htest % (m1image,logfile))
                   verbex(jtest % (m1image,logfile))
                   if science or lamp:
                      verbex(etest % (orimage,csect1a-1,csect1b,o1image,logfile))
                      verbex(htest % (o1image,logfile))
                      verbex(jtest % (o1image,logfile))
                   if nns:
                      verbex(etest % (srimage2,csect1a-1,csect1b,s1image2,logfile))
                      verbex(htest % (s1image2,logfile))
                      verbex(jtest % (s1image2,logfile))
                      verbex(etest % (mrimage2,csect1a-1,csect1b,m1image2,logfile))
                      verbex(htest % (m1image2,logfile))
                      verbex(jtest % (m1image2,logfile))
                      if science or lamp:
                         verbex(etest % (orimage2,csect1a-1,csect1b,o1image2,logfile))
                         verbex(htest % (o1image2,logfile))
                         verbex(jtest % (o1image2,logfile))

               os.chdir("../")

    if redux:
       eimages1 = ["%s/%s_%sc%d%s.fits"%(Sroot,root,objid,chip,(flatfield and "sf" or "")) for chip in chips]
       echips1 = [chip for chip,eimage1 in zip(chips,eimages1) if os.path.exists(eimage1)]
       eimages1 = [eimage1 for chip,eimage1 in zip(chips,eimages1) if os.path.exists(eimage1)]
       echips2 = []
       eimages2 = []
       if nns:
          eimages2 = ["%s/%s_%sc%d%s.fits"%(Sroot,root2,objid,chip,(flatfield and "sf" or "")) for chip in chips]
          echips2 = [chip for chip,eimage2 in zip(chips,eimages2) if os.path.exists(eimage2)]
          eimages2 = [eimage2 for chip,eimage2 in zip(chips,eimages2) if os.path.exists(eimage2)]

       for echips,eimages,eroot in zip([echips1,echips2],[eimages1,eimages2],[root,root2]):
          if len(eimages):
             s1images = [eimage.replace(".fits","r01.fits") for eimage in eimages]
             if science or lamp:
                m1images = [eimage.replace(".fits","mr01.fits") for eimage in eimages]
                o1images = [eimage.replace(".fits","or01.fits") for eimage in eimages]
             elif flat:
                m1images = [eimage.replace(".fits","mr01.fits") for eimage in eimages]
                o1images = []

             Fs = [FITS(s1image) for s1image in s1images]
             naxis1s = array([F["naxis1"] for F in Fs])
             naxis2s = array([F["naxis2"] for F in Fs])
             csectas = array([F["csect1a"] for F in Fs])
             csectbs = array([F["csect1b"] for F in Fs])
             objposs = array([F["objpos"] for F in Fs])
             [F.close() for F in Fs]
             print objposs
             vo = objposs-objposs[argmax(naxis1s)]
             print vo
             minj = echips[argmin(naxis2s)]
             min2 = min(naxis2s)
             smin2 = ":%d" % (min2)
#             if minj in [1,2,3,4]:
#                smin2 = "%d:" % (-min2)
#             elif minj in [5,6,7,8]:
#                smin2 = ":%d" % (min2)

             s1 = "%s/%s_%srsum01.fits"%(Sroot,eroot,objid)
             m1 = "%s/%s_%smrsum01.fits"%(Sroot,eroot,objid)

             vshifts = "+".join(["VTKImageShift(i%d,0,%s,numret=1)[%s]" % (1+i,vo[i],smin2) for i in range(len(vo))])
             verbex(" ".join(["efits"]+s1images+["\"%s\""%(vshifts)]+[s1]))

             verbex(" ".join(["efits"]+m1images+["\"%s\""%(vshifts)]+[m1]))

#             verbex(" ".join(["efits"]+m1images+["\"sum([a[%s] for a in allimages])\""%(smin2)]+[m1]))

             if science or lamp:
                o1 = "%s/%s_%sorsum01.fits"%(Sroot,eroot,objid)
                verbex(" ".join(["efits"]+o1images+["\"%s\""%(vshifts)]+[o1]))

#                verbex(" ".join(["efits"]+o1images+["\"sum([a[%s] for a in allimages])\""%(smin2)]+[o1]))

#    os.remove("%s/%s.pck"%(Sroot,objid))

    return 1

def GetAps(o1,inter=0):
    if not isinstance(o1,StringType): o1 = o1[0]
    if inter: print "Opening",o1
    nns = False
    F = FITS(o1)
    d1 = F.data()
    dweltime = F["dweltime"]
    ncycles = F["ncycles"]
    nshuffle = F["nshuffle"]
    if dweltime and ncycles and nshuffle:
       nns = True
       nodr = F["nod-ra"]
       nodd = F["nod-dec"]
       nod = int(sqrt(nodr**2+nodd**2)/ARCS_PIX+0.5)
       od = os.path.dirname(o1)
       ob = os.path.basename(o1)
       o2 = "%s/%s" % (od,"_".join([ob.split("_")[0][:-1]+"s"]+ob.split("_")[1:]))
       if inter: print "Opening",o2
       F2 = FITS(o2)
       d2 = F2.data()
       F2.close()
    F.close()
    m1,s1 = transpose([bwt(compress(i1,i1)) for i1 in d1])
    s1 = s1/sqrt(d1.shape[1])
#    s1 = divz(m1,s1/sqrt(d1.shape[1]))
    if inter:
       qdump("/tmp/m1.fits",m1)
       qdump("/tmp/s1.fits",s1)
    if nns:
       m2,s2 = transpose([bwt(compress(i2,i2)) for i2 in d2])
       s2 = s2/sqrt(d2.shape[1])
#       s2 = divz(m2,s2/sqrt(d2.shape[1]))
       m2 = shiftim(m2,-nod)
       s2 = shiftim(s2,-nod)
       m1 = m1 + m2
       s1 = sqrt(s1**2 + s2**2)/sqrt(2)
       if inter:
          qdump("/tmp/m2.fits",m2)
          qdump("/tmp/s2.fits",s2)
          qdump("/tmp/m3.fits",m1)
          qdump("/tmp/s3.fits",s1)
    x1 = 1+arange(m1.shape[0])
    n = len(m1)
    w1 = hamming(len(x1))
    g1 = [x1[argmax(m1[len(m1)/4:len(m1)*3/4])+len(m1)/4],log(2.0)]
    sol = GaussFit(g1,x1,m1,divz(w1,s1))
    a1 = Gauss1(sol[0],x1,m1,divz(w1,s1),amp=1)
    x2 = sum(Gauss1(sol[0],x1,m1,divz(w1,s1),amp=0)**2)
    sol[0][1] = exp(sol[0][1])+1.0
    if inter:
       print sol
       print a1,x2
       print diag(sol[1])
       raw_input("<cr>")
    return sol,a1

def Gauss1(g,x,y,w,amp=0,verb=0):
    p = exp(-0.5*(x-g[0])**2/(1+exp(g[1]))**2)
    a = divz(sum((p*y)*w*w),sum((p*p)*w*w))
    p = a*p
    r = w*(p-y)
    if verb: print g[0],1+exp(g[1]),a,sum(r**2)
    if amp: return a
    else: return r

def GaussFit(g,x,y,w):
    sol = leastsq(Gauss1,g,(x,y,w),factor=0.1,ftol=1e-4,xtol=1e-4,full_output=1)
    return sol
    
def ReadWFM(wfm):
   S = {}
   wf = open(wfm,"rb")
   cload = cPickle.load(wf)
   print len(cload)
   if len(cload) >= 11:
      [wCord,wWord,wSord,wave1,wave2] = cload[:5]
      [wCord,wWord,wSord,wave1,wave2,wxsol,wxrms,wxbwt,wysol,wyrms,wybwt] = cload[:11]
      nC = sum([1 for j in range(1+wCord) for i in range(1+wCord-j)])
      nW = wWord +1
#      gl = len(wysol) == nC*nW
      S[0] = cload[5:11]
      gl = True
      arots = []
      if len(cload) > 11:
         for chip in 1+arange(8):
             print chip,
             S[chip] = cload[5+chip*6:5+(chip+1)*6]
             sys.stdout.flush()
         print
         gl = False

#      if len(cload) == 12:
#         arots = cload[11]
#         if arots.any(): gl = False
#   else:
#      [wCord,wWord,wSord,wave1,wave2] = cload[:5]
#      S[0] = cload[5:5+6]
#      for chip in 1+arange(8):
#          print chip,
#          S[chip] = cload[5+chip*6:5+(chip+1)*6]
#          sys.stdout.flush()
#      print
#      [wxsol,wxrms,wxbwt,wysol,wyrms,wybwt] = S[0]
   wf.close()
   print gl,wCord,wWord,wSord,wave1,wave2,len(wxsol),len(wysol),wxrms/MINS_PIX,wyrms/MINS_PIX
   Cord = wCord
   Word = wWord
   Sord = wSord
   return gl,wave1,wave2,Cord,Word,Sord,S,arots


