
import time
import signal
import sys
import os
import threading
import Queue
import basics
from multiprocessing import Process, Pool, Manager
from multiprocessing import log_to_stderr
import logging
from TwistedHub import reactor
from TwistedHub import IMPORTS
from TwistedHub import twisted_import as multiprocimport
import cStringIO, cPickle, marshal, traceback

logger = log_to_stderr()

if os.environ.has_key('PYTHONTHREADS'):
    MaxThreads = int(os.environ['PYTHONTHREADS'])
else:
    MaxThreads = 1

if os.environ.has_key('PYTHONTHREADOUT'):
    ThreadOutput = int(os.environ['PYTHONTHREADOUT'])
else:
    ThreadOutput = 0

class Worker(threading.Thread):
  def __init__(th,queue,task,listing=None):
     th.__queue = queue
     th.__task = task
     th.__listing = listing
     threading.Thread.__init__(th)
  def run(th):
     while 1:
        item = th.__queue.get()
        if item is None: break
        if th.__listing:
           if not ThreadOutput:
              sys.stdout = open("/dev/null","w")
              sys.stderr = sys.stdout
           th.__task(th.__listing[item])
           if not ThreadOutput:
              sys.stdout.close()
              sys.stdout = sys.__stdout__
              sys.stderr = sys.__stderr__
        else:
           if not ThreadOutput:
               sys.stdout = open("/dev/null","w")
               sys.stderr = sys.stdout
           th.__task()
           if not ThreadOutput:
               sys.stdout.close()
               sys.stdout = sys.__stdout__
               sys.stderr = sys.__stderr__

def RunThreadedQueue(func,listing,MaxThreads=MaxThreads):
    print "Setting up with a maximum of %d threads." % (MaxThreads)
    workers = min([MaxThreads,len(listing)])
    queue = Queue.Queue(MaxThreads)
    for item in range(workers):
        Worker(queue,func,listing).start()
    for j in range(len(listing)):
        queue.put(j)
    for item in range(workers):
        queue.put(None)
    while threading.activeCount() > 1: pass
    return queue

def RunTheFork(func,args,ThreadOutput=ThreadOutput):
    pid = os.getpid()
    pidfile = "/tmp/pid%d.lock" % (pid)
    lock = open(pidfile,"w")
    if not ThreadOutput:
       out = open(os.path.basename(pidfile).replace(".lock",".out"),"w")
       sys.stdout = out
       sys.stderr = out
    value = apply(func,(args,))
    lock.close()
    if not ThreadOutput:
       out.close()
       os.remove(out.name)
       sys.stdout = sys.__stdout__
       sys.stderr = sys.__stderr__
    os.remove(pidfile)
    time.sleep(0.1)
    return value

class ForkFunc: 
    def __init__(self,func):
       self.func = func
    def run(self,args,ThreadOutput=ThreadOutput):
       p = Process(target=RunTheFork,args=(self.func,args),kwargs={"ThreadOutput":ThreadOutput})
       p.start()
       return p
    def oldrun(self,args,ThreadOutput=ThreadOutput):
       child = os.fork()
       if child == 0:
           try:
              RunTheFork(self.func,args,ThreadOutput=ThreadOutput)
              status = 0
           except:
              status = -1
           os._exit(status)
       else:
           return child

def RunForkedQueue(func,listing,sleeptime=0.1,MaxThreads=MaxThreads,monitor=1,verb=1,ThreadOutput=ThreadOutput):
    CHILDREN = []
    returncodes = []
    print "Setting up with a maximum of %d forks." % (MaxThreads)
    workers = min([MaxThreads,len(listing)])
    fl = ForkFunc(func)
    iob = 0
    tob = 0
    while iob < len(listing) or len(CHILDREN) > 0:
       if len(CHILDREN) < MaxThreads and iob < len(listing):
          if verb: print "Starting:",listing[iob]
          child = fl.run(listing[iob],ThreadOutput=ThreadOutput)
          print "New Child=",child
          CHILDREN.append(child)
          iob += 1
       elif len(CHILDREN) > 0:
          DONE = []
          for child in CHILDREN:
             pid = child.pid
             if not os.path.exists("/tmp/pid%d.lock" % (pid)): DONE.append(child)
             elif monitor and tob and not (tob%50) : basics.verbex("ps u -p %d" % (pid))
          for child in DONE:
             if not child.is_alive():
                pid = child.pid
                print "Removing %d." % (pid)
                CHILDREN.remove(child)
                stat = child.exitcode
                returncodes.append(os.WIFEXITED(stat))
       time.sleep(sleeptime)
       tob += 1
#       print CHILDREN

    while len(CHILDREN) > 0:
       time.sleep(2*sleeptime0)
       DONE = []
       for pid in CHILDREN:
           if not os.path.exists("/tmp/pid%d.lock" % (pid)): DONE.append(pid)
           elif monitor and tob and not (tob%50) : basics.verbex("ps u -p %d" % (pid))
       for pid in DONE: CHILDREN.remove(pid)
    print returncodes
    return returncodes

def quietMinitask(fa):
    func,a = fa
    v = eval(func)(a)
    return v

def verbMinitask(fa):
    if os.path.exists("pid%d.out"%(os.getpid())):
       p = open("pid%d.out"%(os.getpid()),"a")
    else:
       p = open("pid%d.out"%(os.getpid()),"w")
    func,a = fa
    if not globals().has_key(func):
       logger.warning("%s not in globals()" % (func))
    try:
      sys.stdout = p
      sys.stderr = p
      v = eval(func)(a)
      p.close()
      sys.stdout = sys.__stdout__
      sys.stderr = sys.__stderr__
#      os.remove(p.name)
    except:
      v = traceback.print_exc()
      logger.warning(v)
      p.write(v)
      p.close()
    return v

Minitask = quietMinitask

def RunMinitasks(func,args,namespace,NP=int(os.environ.get("NP",1)),chunksize=8,verb=0):
    [globals().setdefault(k,v) for k,v in namespace.items()]
    logger.warning('STARTING %s'%(func.__name__))
    if verb:
       logger.setLevel(logging.INFO)
    fa = [(func.__name__,a) for a in args]
    if reactor.running:
       return map(Minitask,fa)
    else:
       po = Pool(NP)
       vo = po.map_async(Minitask,fa,chunksize)
       po.close()
       po.join()
       vo.wait()
       return vo.get()

if __name__ == "__main__":
   def TestFunc(w):
       sys.stdout.write("PID=%d: %s\n" % (os.getpid(),w[0]))
       return w
   args = [("Hello %d\n"%(j),) for j in range(100)]
   ns = globals()
   NP = int(os.environ.get("NP",0))
   if NP:
     RunMinitasks(TestFunc,args,ns,NP=NP)
   else:
     map(TestFunc,args)


